from typing_extensions import Never
import requests, json, time, pytz
from calendar import monthrange
from datetime import datetime, timedelta
from .interfaces import (
    BRAND_CODE,
    SUBCATEGORY_CODE,
    CalendarResponse,
    validate_subcategory_code,
    validate_brand_code,
)


def validate_brands_and_catogories(
    brands: list[BRAND_CODE], subcategories: list[SUBCATEGORY_CODE]
):
    for brand in brands:
        validate_brand_code(brand)
    for subcategory in subcategories:
        validate_subcategory_code(subcategory)


def fetch_news_for_today():
    return NotImplemented


def fetch_schedule_for_today(
    brands: list[BRAND_CODE] = [],
    subcategories: list[SUBCATEGORY_CODE] = [],
):
    validate_brands_and_catogories(brands, subcategories)
    japan_timezone = pytz.timezone("Asia/Tokyo")
    now_in_japan = datetime.now(japan_timezone)

    midnight_today = japan_timezone.localize(
        datetime(now_in_japan.year, now_in_japan.month, now_in_japan.day, 0, 0, 0)
    )
    midnight_tomorrow = midnight_today + timedelta(hours=23, minutes=59, seconds=59)
    return _fetch_schedule(
        midnight_today,
        midnight_tomorrow,
        brands,
        subcategories,
    )


def fetch_schedule_for_month(
    year: int = 0,
    month: int = 0,
    brands: list[BRAND_CODE] = [],
    subcategories: list[SUBCATEGORY_CODE] = [],
):
    validate_brands_and_catogories(brands, subcategories)
    japan_timezone = pytz.timezone("Asia/Tokyo")

    if year == 0:
        year = datetime.now(japan_timezone).year

    if month == 0:
        month = datetime.now(japan_timezone).month

    first_day_this_month = japan_timezone.localize(datetime(year, month, 1, 0, 0, 0))
    last_day = monthrange(year, month)[1]
    last_second_this_month = japan_timezone.localize(
        datetime(year, month, last_day, 23, 59, 59)
    )

    return _fetch_schedule(
        first_day_this_month,
        last_second_this_month,
        brands,
        subcategories,
    )


# Not Recommended to use because arbitary api call may not occur from idolmaster portal
# This interface is for internal use
def _fetch_schedule(
    start_date: datetime,
    end_date: datetime,
    brands: list[BRAND_CODE] = [],
    subcategories: list[SUBCATEGORY_CODE] = [],
    limit=200,
) -> CalendarResponse:
    base_url = "https://cmsapi-frontend.idolmaster-official.jp/sitern/api/idolmaster/Article/list"

    target_start_date = start_date.astimezone(pytz.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
    target_end_date = end_date.astimezone(pytz.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

    params = {
        "site": "jp",
        "ip": "idolmaster",
        "token": _get_token(),
        "sort": "asc",
        "limit": limit,
        "data": {
            "category": ["SCHEDULE"],
            "brand": brands,
            "subcategory": subcategories,
            "target_start_date": target_start_date,
            "target_end_date": target_end_date,
        },
    }

    params["data"] = json.dumps(params["data"])

    headers = {
        "accept": "application/json, text/plain, */*",
        "accept-language": "ja",
        "cache-control": "no-cache",
        "pragma": "no-cache",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-site",
    }

    response = requests.get(base_url, headers=headers, params=params)

    if response.status_code == 200:
        return response.json()["data"]
    else:
        response.raise_for_status()
        return Never


def _get_token() -> str:
    time.sleep(0.1)
    # Define the URL
    url = "https://cmsapi-frontend.idolmaster-official.jp/sitern/api/cmsbase/Token/get"

    # Set up the headers as given in the fetch request
    headers = {
        "accept": "application/json, text/plain, */*",
        "accept-language": "ja",
        "sec-ch-ua": '"Google Chrome";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"Windows"',
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-site",
    }

    # Additional options from fetch request translated to requests library
    cookies = {
        # Cookies must be set if needed, fetch has "credentials": "include"
    }
    # In requests, referrer and referrerPolicy do not have direct equivalents,
    # but can be added to headers if necessary.

    # Send the GET request
    response = requests.get(url, headers=headers, cookies=cookies)

    # Print the response text (or process it in other ways)
    return response.json()["data"]["token"]



import io, csv
from typing import Union, TypedDict


class DialogueLine(TypedDict):
    id: str
    name: str
    text: str
    trans: str


class StoryCsv:
    def __init__(self, csv_text: Union[str, list[str]]):
        if isinstance(csv_text, str):
            csv_text.replace("\r\n", "\n")
            csv_text = csv_text.strip().split("\n")

        reader = csv.DictReader(csv_text)
        # remove the starting line
        if reader.fieldnames != ["id", "name", "text", "trans"]:
            raise ValueError(
                f"First line of csv should be 'id,name,text,trans'"
            )

        self.data: list[DialogueLine] = []
        for line in reader:
            self.append_line(
                {
                    "id": line["id"],
                    "name": line["name"],
                    "text": line["text"],
                    "trans": line["trans"],
                }
            )
        
        translator_line = self.data.pop(-1)
        if translator_line["id"] != "译者":
            raise ValueError(
                f"Last line of csv should be translator and start with '译者'"
            )
        self.translator = translator_line["name"]
        
        info_line = self.data.pop(-1)
        if info_line["id"] != "info":
            raise ValueError(
                f"Last but 2 line of csv should be original file info and start with 'info'"
            )
        self.origin = info_line["name"]

    @classmethod
    def new_empty_csv(cls, full_origin: str) -> "StoryCsv":
        return cls(
            f"id,name,text,trans\ninfo,{full_origin},,\n译者,,,"
        )
    
    def append_line(self, line: DialogueLine):
        self.data.append(line)

    def __str__(self):
        with io.StringIO() as output:

            writer = csv.DictWriter(output, fieldnames=["id", "name", "text", "trans"])

            # write header
            writer.writerow(
                {"id": "id", "name": "name", "text": "text", "trans": "trans"}
            )

            for row in self.data:
                writer.writerow(row)
            writer.writerow(
                {"id": "info", "name": self.origin, "text": "", "trans": ""}
            )
            writer.writerow(
                {"id": "译者", "name": self.translator, "text": "", "trans": ""}
            )

            csv_string = output.getvalue()
            return csv_string.replace("\r\n", "\n")



import requests, json, re
from typing import Dict, Any, TypedDict, Union
from urllib.parse import quote
from bs4 import BeautifulSoup

class LyricMetadata(TypedDict):
    title: str
    artist: str
    composer: str
    lyricist: str
    lyrics: str

def is_valid_json(json_str):
    try:
        # 使用json.loads尝试解析字符串
        json.loads(json_str)
        return True
    except json.JSONDecodeError:
        # 如果解析失败，返回False
        return False


def find_json_object_in_string(s: str):

    start_c = "{"
    end_c = "}"

    start = s.find(start_c)
    end = s.rfind(end_c)

    if end == -1 or start >= end:
        raise Exception("No matching closing bracket found")

    if is_valid_json(s[start : end + 1]):
        return s[start : end + 1]

    return find_json_object_in_string(s[start:end])


def query_song(song_name: str) -> Dict[str, Any]:  # type: ignore
    res = requests.get(
        f"https://recochoku.jp/search/song?q={quote(song_name)}&type=SINGLE",
    )
    soup = BeautifulSoup(res.content, "html.parser")
    scripts = soup.find_all("script")

    keyword = "var musicList ="

    # 遍历所有 script 标签并打印其中的文本内容：
    for script in scripts:
        if keyword in script.get_text():
            for line in script.get_text().split("\n"):
                if keyword in line:
                    return json.loads(find_json_object_in_string(line))


def query_track_id(music_id) -> int:
    res = requests.get(
        f"https://recochoku.jp/song/S{music_id}",
    )
    soup = BeautifulSoup(res.content, "html.parser")
    scripts = soup.find_all("script")
    keyword = "trackId = "
    for script in scripts:
        if keyword in script.get_text():
            for line in reversed(script.get_text().split("\n")):
                if keyword in line:
                    pattern = r"\d+"

                    matches = re.findall(pattern, line)
                    return int(matches[0])
    raise Exception("Track ID not found")


def query_metadata_from_api_store(track_id):
    res = requests.get(
        f"https://common-api-store.recochoku.jp/lyrics/{track_id}",
        headers={
            "accept": "application/json, text/javascript, */*; q=0.01",
            "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7,ja;q=0.6,zh-TW;q=0.5",
            "sec-ch-ua": '"Chromium";v="122", "Not(A:Brand";v="24", "Google Chrome";v="122"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-site",
            "x-sec-ch-ua-platform-version": "15.0.0",
            "Referer": "https://recochoku.jp/",
            "Referrer-Policy": "strict-origin-when-cross-origin",
        },
    )
    tmp = res.json()
    composer = tmp.get("composer")
    lyricist = tmp.get("lyricist")
    artist = tmp.get("artist")
    lyrics = get_lyrics_from_words(tmp.get("words"))
    return {
        "composer": composer,
        "lyricist": lyricist,
        "artist": artist,
        "lyrics": lyrics,
    }


def get_lyrics_from_words(words: Union[str, None]):
    if words is None:
        return ""
    rtn = ""
    for word in json.loads(words):
        rtn += chr(word)
    return rtn


def query_metadata_for(music_name: str, choice: int = 0) -> LyricMetadata:
    rtn = query_song(music_name)
    if choice > len(rtn["music"]):
        raise Exception(f"Choice index out of range: only {choice} choices are returned")
    music_id = rtn["music"][choice]["musicId"]
    track_name = rtn["music"][choice]["title"]
    artist_name = rtn["music"][choice]["artist"]["name"]
    track_id = query_track_id(music_id)
    metadata = query_metadata_from_api_store(track_id)
    return {
        "title": track_name,
        "artist": artist_name,
        "composer": metadata["composer"],
        "lyricist": metadata["lyricist"],
        "lyrics": metadata["lyrics"],
    }



from typing import Callable, Optional, Union
from .gakuen_parser import parse_messages, encode_messages
from .story_csv import StoryCsv


def gakuen_txt_to_sc_csv(gakuen_txt: str, txt_name_without_ext: str) -> str:
    parsed = parse_messages(gakuen_txt)
    sc_csv = StoryCsv.new_empty_csv(f"{txt_name_without_ext}.txt")
    for line in parsed:
        if line["__tag__"] == "message":
            if line.get("text") is not None and line.get("name") is not None:
                sc_csv.append_line(
                    {
                        "id": "0000000000000",
                        "name": line.get("name"),
                        "text": line.get("text"),
                        "trans": "",
                    } # type: ignore
                )
        if line["__tag__"] == "choicegroup":
            if isinstance(line["choices"], list):
                for choice in line["choices"]:
                    sc_csv.append_line(
                        {
                            "id": "select",
                            "name": "",
                            "text": choice["text"],
                            "trans": "",
                        }
                    )
            elif isinstance(line["choices"], dict):
                sc_csv.append_line(
                    {
                        "id": "select",
                        "name": "",
                        "text": line["choices"]["text"],
                        "trans": "",
                    }
                )
            else:
                raise ValueError(f"Unknown choice type: {line['choices']}")

    return str(sc_csv)


def trivial_translation_merger(
    original_text: str,
    translated_text: str,
    validation_original_text: Optional[str] = None,
):
    if (
        validation_original_text is not None
        and validation_original_text != original_text
    ):
        raise ValueError(
            f"Original text does not match validation text: {validation_original_text} != {original_text}"
        )
    return translated_text


# eg <r\=はなみさき>花海咲季</r>
def line_level_dual_lang_translation_merger_in(
    original_text: str,
    translated_text: str,
    validation_original_text: Optional[str] = None,
):
    raise NotImplementedError


def merge_translated_csv_into_txt(
    csv_text: Union[str, list[str]],
    gakuen_txt: str,
    merger: Callable[[str, str, Optional[str]], str],
) -> str:
    story_csv = StoryCsv(csv_text)
    parsed = parse_messages(gakuen_txt)
    iterator = iter(story_csv.data)

    for line in parsed:
        if line["__tag__"] == "message":
            if line.get("text"):
                next_csv_line = next(iterator)
                line["text"] = merger(
                    line["text"], next_csv_line["trans"], next_csv_line["text"]
                )
        if line["__tag__"] == "choicegroup":
            if isinstance(line["choices"], list):
                for choice in line["choices"]:
                    next_csv_line = next(iterator)
                    choice["text"] = merger(
                        choice["text"], next_csv_line["trans"], next_csv_line["text"]
                    )
            elif isinstance(line["choices"], dict):
                next_csv_line = next(iterator)
                line["choices"]["text"] = merger(
                    line["choices"]["text"],
                    next_csv_line["trans"],
                    next_csv_line["text"],
                )
    return encode_messages(parsed)



from typing import List, Optional, Any, Literal, get_args, Type
from typing_extensions import TypedDict

BRAND_CODE = Literal[
    "IDOLMASTER", "CINDERELLAGIRLS", "MILLIONLIVE", "SIDEM", "SHINYCOLORS", "GAKUEN", "OTHER"
]

SUBCATEGORY_CODE = Literal[
    "GAME",
    "EVENT",
    "ANIME",
    "LIVESTREAM",
    "RADIO",
    "GOODS",
    "OTHER",
    "DVD-BD",
    "BOOK",
    "CD",
    "COLLABO-CAMP",
]

class InvalidBrandCode(ValueError):
    pass


class InvalidSubcategoryCode(ValueError):
    pass


def validate_brand_code(value: str):
    if is_literal_str(value, BRAND_CODE):
        return
    raise InvalidBrandCode(f"Invalid brand code: {value}")

def validate_subcategory_code(value: str):
    if is_literal_str(value, SUBCATEGORY_CODE):
        return
    raise InvalidSubcategoryCode(f"Invalid subcategory code: {value}")

def is_literal_str(value: str, typ) -> bool:
    return isinstance(value, str) and value in get_args(typ)


class Brand(TypedDict):
    name: str
    code: BRAND_CODE


class Subcategory(TypedDict):
    name: str
    code: SUBCATEGORY_CODE


class Categories(TypedDict):
    name: str
    subcategory: List[Subcategory]
    code: str


class RepeatPattern(TypedDict, total=False):
    date: int
    repeat_startdate: int
    kind: str
    days: List[int]
    repeat_pattern: List[Any]
    repeat_enddate: int
    exceptions: List[int]


class ArticleList(TypedDict, total=False):
    content: str
    information: str
    event_enddate: int
    event_url: str
    startdate: int
    branch_no: int
    _p_key: str
    site: str
    _id: str
    region: str
    path: str
    display_state_during_period: str
    dspdate: Any
    ip: str
    brand: List[Brand]
    event_startdate: int
    listed_subcategories: str
    repeat: bool
    repeat_pattern: RepeatPattern
    updated: int
    enddate: int
    publish_status: str
    categories: Categories
    delflg: str
    event_dspdate: str
    title: str
    event_place: Any
    news_thumbnail: Any
    hashtag: Any
    thumbnail: str
    filename: str
    sp_thumbnail: str
    url: str


class CalendarResponse(TypedDict):
    apiStatus: bool
    total_count: int
    start: int
    limit: int
    count: int
    time: int
    article_list: List[ArticleList]



from typing import Any

def parse_identifier(message, i):
    j = i
    while message[i].isalnum() or message[i] == "_":
        i = i + 1
    return i, message[j:i]


def parse_string(message, i):
    j = i
    n = 0
    while n != 0 or (message[i] != "=" and message[i] != "]"):
        if message[i] == "\\":
            i = i + 1
            if message[i] == "{":
                n = n + 1
            elif message[i] == "}":
                n = n - 1
                if n < 0:
                    raise ValueError()
        i = i + 1
    if message[i] == "=":
        i = i - 1
        while message[i] != " ":
            i = i - 1
    return i, message[j:i]


def parse_object(message, i):
    if message[i] != "[":
        raise ValueError()
    i = i + 1
    i, tag = parse_identifier(message, i)
    obj = {"__tag__": tag}
    while message[i] != "]":
        if message[i] != " ":
            raise ValueError()
        i = i + 1
        i, key = parse_identifier(message, i)
        if key == "__tag__":
            raise ValueError()
        if message[i] != "=":
            raise ValueError()
        i = i + 1
        if message[i] == "[":
            i, value = parse_object(message, i)
        else:
            i, value = parse_string(message, i)
        if key not in obj:
            obj[key] = value
        elif type(obj[key]) is list:
            obj[key].append(value)
        else:
            obj[key] = [obj[key], value]
    if message[i] != "]":
        raise ValueError()
    i = i + 1
    return i, obj


def parse_message(message):
    i, o = parse_object(message, 0)
    if i != len(message):
        raise ValueError()
    return o


def parse_messages(messages: str) -> list[dict[str, Any]]:
    return [
        parse_message(message) for message in messages.split("\n") if message.strip()
    ]


def encode_message_kv(k, v):
    if k == "__tag__":
        if type(v) is not str:
            raise ValueError()
        return v
    if type(v) is list:
        return " ".join([encode_message_kv(k, w) for w in v]) # type: ignore
    if type(v) is str:
        return f"{k}={v}"
    if type(v) is not dict:
        return ValueError()
    if "__tag__" not in v:
        raise ValueError()
    return f"{k}={encode_message(v)}"


def encode_message(obj):
    return (
        "["
        + " ".join(
            [obj["__tag__"]]
            + [encode_message_kv(k, v) for k, v in obj.items() if k != "__tag__"]
        )
        + "]"
    )


def encode_messages(objs) -> str:
    return "\n".join(encode_message(obj) for obj in objs)
